/*
	builds a makefile
	
	T.AL 95
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tos.h>

#if !defined(FMAKE_H)
#include "fmake.h"
#endif
#if !defined(TOOLS_H)
#include "tools.h"
#endif


int ind;
char spaces[256];

void indent(int i)
{
	char *spc;
	
	for(spc= spaces; i>0; i--) *spc++= ' '; 
	*spc= 0;
}

void display_fname(File *file, int found)
{
	if(found)
		printf("%s", spaces);
	else
		printf("%snew ", spaces);
		
	if(file->fname)
		printf("file: %s", file->fname);
	else
		printf("file: <not defined>");
		
	if(found)
	{
		printf(" %02d/%02d/%04d %02d:%02d:%02d\n", 
		         file->date.day, file->date.month, file->date.year,
		         file->date.h, file->date.mn, file->date.sec);
	}
	else
		putchar('\n');
}

void display_env(VAR *var)
{
	if(var)
	{
		while(var)
		{
			if(var->ext)
				printf("%senv: %s%s= `%s`\n", spaces, var->name, var->ext, var->value);
			else if(var->value)
				printf("%senv: %s= `%s`\n", spaces, var->name, var->value);
			else
				printf("%senv: %s\n", spaces, var->name);
				
			var= var->next;
		}
	}
	else
		printf("%senv: <not defined>\n", spaces);
}

void display_a_cmd(CMD *cmd)
{
	ARG *arg;
	
	if(cmd)
	{
		printf("%s'%s'", spaces, cmd->fname);
		arg= cmd->argv;
		while(arg)
		{
			printf(" `%s`", arg->name);
			if(arg->value)
				printf("%s", arg->value);
			arg= arg->next;
		}
		putchar('\n');
	}
	else
		printf("%scmd: <not defined>\n", spaces);
}

void display_cmd(CMD *cmd)
{
	ARG *arg;
	
	if(cmd)
	{
		while(cmd)
		{
			printf("%scmd: %s", spaces, cmd->fname);
			arg= cmd->argv;
			while(arg)
			{
				printf(" `%s`", arg->name);
				if(arg->value)
					printf("%s", arg->value);
				arg= arg->next;
			}
			putchar('\n');
			cmd= cmd->next;
		}
	}	
	else
		printf("%scmd: <not defined>\n", spaces);
}


CMD *find_command(char *outext, char *inext, CMD *cmd)
{
	if(outext && inext)
		for(; cmd 
		   && ((cmd->outext && strcmp(cmd->outext, outext) || !cmd->outext) || 
		      ((cmd->inext &&  strcmp(cmd->inext,  inext) || !cmd->inext)))
		    ; cmd= cmd->next) ;

	else if(!outext && !inext)
		for(; cmd && (cmd->outext || cmd->inext); cmd= cmd->next) ;
	
	else
		alert("fmake: can't find command for building `*.%s` with `*.%s`\n", outext, inext);	
	
	return cmd;
}

/* verifie que les fichiers de dependance de file sont constructibles 
   avec les commandes disponibles
*/
int check_flist(File *file, CMD *cmd)
{
	char *outext;
	char *inext;
	File *f;
	
	/* recupere l'extension de target */ 
	outext= getext(file->fname);
	
	for(f= file->flist; f; f= f->next, free(inext))
	{
		/* recupere l'extension de source */ 
		inext= getext(f->fname);

		/* recherche une commande utilisable */
		if(!find_command(outext, inext, cmd))
			break;
	}

	free(outext); free(inext);
	return f!=NULL;
}


VAR *find_var(char *name, char *ext, VAR *var)
{
	while(var)
	{
		if( ( ext && var->ext && strcmp(var->ext, ext)==0
		    || !ext && !var->ext)
		&&  ( name && strcmp(var->name, name)==0
		    || !name))
			return var;
		else
			var= var->next;
	}

	return NULL;
}

/* construit target
   renvoie :
   FATAL en cas d'erreur
   UPDATE si construction possible
   UP_TO_DATE si pas de commande definie
*/
int build_target(File *target, File *source)
{
	int status= 0;
	char *outext;
	char *inext;
	CMD *cmd;

	outext= getext(target->fname); /* extension de target */ 
	inext=  getext(source->fname); /* extension de source */
	cmd= find_command(outext, inext, target->cmd);

	if(!cmd) return UP_TO_DATE;

	printf("building `%s` with `%s`\n", basename(target->fname), basename(source->fname));

	for(; cmd && !status; 
	      cmd= find_command(outext, inext, cmd->next))
	{
		/* display_a_cmd(cmd); */
		status= exec_cmd(cmd, target, source);
		if(status)
		{
			printf("can't build target `%s` \n", target->fname);
			Fdelete(target->fname);	/* detruit la cible pour forcer 
			          la reconstruction du projet au prochain appel */
		}
	}

	if(status)
		return FATAL;
	else
		return UPDATE;
}


/* construit un projet */
int build_project(File *prj)
{
	int status, fatal= 0;
	CMD *cmd= find_command(NULL, NULL, prj->cmd);
	
	if(!cmd) return UP_TO_DATE;	
	printf("\n\nbuilding project `%s`\n", basename(prj->fname));
	for(; cmd; cmd= find_command(NULL, NULL, cmd->next))
	{
		/* display_a_cmd(cmd); */
		status= exec_cmd(cmd, prj, prj->flist);
		fatal|= status;
	}

	if(fatal)
		printf("game over. insert coins\n");
	else
		printf("yeahh !!! level sucessfully completed !\n");

/*
	printf("\nFMAKE env:\n");
	display_env(prj->env);
	printf("TOS env:\n");
	display_env(tosenv);
*/
	
	return fatal;
}


/* localise un fichier :
	si le chemin d'acces n'est pas fourni
		recherche dans le repertoire courant
		recherche d'un chemin d'acces dans l'environnement du fichier 
		(si une variable #path est definie)
		-> si les recherches echouent, le fichier n'existe pas

	si le chemin d'acces est fourni
		recherche dans ce repertoire
		-> si la le recherche echoue, le fichier n'existe pas 

	=> si les recherches aboutissent file->fname est mis  jour 
	pour designer le chemin d'acces complet au fichier
*/

int locate_a_file(File *file, VAR *env)
{
	char *tmp, *ext;
	VAR *var;
	char path[256];

	if(file->env)
		env= file->env;
	
	if(strchr(file->fname, '\\'))
	{	/* le chemin d'acces est fourni */
		/* printf("%ssearching file `%s`\n", spaces, file->fname); */

		if(Fsfirst(file->fname, 0) <0)	
			return 1; /* fichier n'existe pas */
		else
		{
			set_file_dtm(file);
			return 0;	/* fichier existe */
		}
	}
	else
	{	/* le chemin d'acces n'est pas fourni */
		/* printf("%ssearching file `%s`\n", spaces, file->fname); */

		if(Fsfirst(file->fname, 0) <0)
		{	/* pas dans le repertoire courant */
			
			/* recherche d'un path specifique */
			ext= getext(file->fname);
			var= find_var("#path", ext, env);
			while(var)
			{
				strcpy(path, var->value);
				strcat(path, file->fname);
				/* printf("%ssearching file `%s`\n", spaces, path); */

				if(Fsfirst(path ,0) ==0)
				{
					free(file->fname);
					file->fname= strdup(path);
					set_file_dtm(file);

					return 0; /* le fichier existe */
				}
				var= find_var("#path", ext, var->next);
			}

			/* recherche d'un path global */
			var= find_var("#path", NULL, env);
			while(var)
			{
				strcpy(path, var->value);
				strcat(path, file->fname);
				/* printf("%ssearching file `%s`\n", spaces, path); */
				
				if(Fsfirst(path ,0) ==0)
				{
					free(file->fname);
					file->fname= strdup(path);
					set_file_dtm(file);

					return 0; /* le fichier existe */
				}
				var= find_var("#path", NULL, var->next);
			}

			return 1; /* le fichier n'existe pas */
		}
		else
		{	/* fichier dans le repertoire courant */
			tmp= file->fname;
			file->fname= (char *)malloc(strlen(Path) + strlen(tmp) +1);
			strcpy(file->fname, Path);
			strcat(file->fname, tmp);
			free(tmp);
			set_file_dtm(file);

			return 0;
		}
	}
}

/*
	construire un fichier:
	si le fichier existe, verifier la date de modification de tous
	ces constituants
	-> si un constituant est plus recent que le fichier, il faut 
	reconstruire toute la branche
	
	si le fichier n'existe pas verifier ces consitutants et le 
	reconstruire
	
	erreur si un des sources du fichier n'existe pas
	
	3 valeurs de retour:
	UPDATE     : mise  jour effectue
	UP_TO_DATE : pas de mise  jour effectue
	FATAL      : erreur, un/plusieurs source(s) manque(nt)
*/

int build_a_file(File *file, File_Date *date)
{
	int newfile, status, up_to_date, force_update, fatal;
	File *f;
	File_Date fdate;
	
	if(!file) return UP_TO_DATE;

	if(newfile= locate_a_file(file, file->env))
	{	/* file n'existe pas */
		if(!file->cmd)
		{
			printf("fatal: source file `%s` doesn't exist\n", file->fname);
			/* file n'existe pas et n'est pas constructible */
			return FATAL; /* file est un source inexistant */
		}
	}
		
	/* display_fname(file, !newfile); */
	
	force_update= 0;
	for(fatal= 0, up_to_date= 1, f= file->flist; 
	    f; 
	    f= f->next)
	{
		/* printf("%s{\n", spaces);
		indent( ind+= 4); */

		datecpy(&fdate, &file->date);
		status= build_a_file(f, &fdate);

		/* force la construction */
		if(Make_All && status!=FATAL) status= UPDATE;

		if(status==UPDATE)
		{
			if(file->cmd)
			{
				if(Make_All || newfile || datecmp(&fdate, &file->date)>0)
				{	/* garde la date de mise  jour */
					datecpy(date, &fdate);
					/* reconstruit file */
					fatal|= build_target(file, f)==FATAL;
				}
				else
				{	/* la cible est  jour */
					status= UP_TO_DATE;
					datecpy(date, &file->date);
				}
			}
			else
			{/* fichier de dependance, reconstruire le pere de file */
				force_update= 1;
				datecpy(date, datemax(&file->date, &fdate));
			}
		}

		up_to_date&= status==UP_TO_DATE;
		fatal|= status==FATAL;
		
		/* force la sortie  la premiere erreur */
		if(Make_Stop==STOP_FIRST && fatal) break;
			
		/* indent( ind-= 4);
		printf("%s}\n", spaces); */
	}

	if(fatal) 
		return FATAL;

	else if(!force_update && up_to_date && datecmp(date, &file->date)>=0)
		return UP_TO_DATE;

	else	
	{	/* renvoie la date qui a provoque la mise  jour */
		datecpy(date, datemax(date, &file->date));
		return UPDATE;
	}
}


int builder(File *prj)
{
	int status, fatal= 0;
	int newfile;
	File_Date date;
	File *single;

	while(prj)
	{
		/* detruit le logfile s'il est unique */
		if(Make_Logsinglefile)
		{
			single= new_file();
			single->fname= filename(Make_Logpath, Make_Logbasename, Make_Logext);
			locate_a_file(single, prj->env);
			newfile= Fcreate(single->fname, 0);
			printf("Warning: single logfile `%s`\n", single->fname);
			Fclose(newfile);
			free_file(single);
		}

		/* determine l'heure du projet s'il exite deja */
		newfile= locate_a_file(prj, prj->env);
		datecpy(&date, &prj->date);
		
		/* go ! */
		status= build_a_file(prj, &date);
		if(status!=FATAL)
		{
			if(newfile || status==UPDATE)
				build_project(prj);
		
			else if(status==UP_TO_DATE)
				printf("project `%s` is up to date\n", basename(prj->fname));
		}	
		else
		{
			fatal= 1;
			printf("there are errors, project `%s` isn't built\n", prj->fname);
		}

		prj= prj->next;
	}
	
	return fatal;
}
