/*
	fmake.c

	builds a project
		
	T.AL 95
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tos.h>

#if !defined(FMAKE_H)
#include "fmake.h"
#endif
#if !defined(TOOLS_H)
#include "tools.h"
#endif

char Path[256];
DTA Dta;

File_Date Make_Date;
int       Make_Stop, 
          Make_All;

char      Make_Logpath[256], Make_Logbasename[12], Make_Logext[4];
int       Make_Logsinglefile;

char      Make_Buildfilespec[256];

VAR *tosenv;



int cleaner(File *file)
{
	int status= 0;
	File *f;
		
	if(!file) return 0;

	if(locate_a_file(file, file->env))
	{	/* file n'existe pas */
		if(!file->cmd)
		{
			printf("fatal: source file `%s` doesn't exist\n", file->fname);
			/* file n'existe pas et n'est pas constructible */
			return FATAL; /* file est un source inexistant */
		}
		else
			printf("already deleted `%s`\n", basename(file->fname));
	}
	else	
		if(file->cmd)
		{
			printf("deleting `%s`\n", basename(file->fname));
			status|= Fdelete(file->fname) <0;
		}

	for(f= file->flist; f; f= f->next)
		status|= cleaner(f);
	
	return status;
}

int touch_file(File *file)
{
	/* int h;
	DOSTIME newdt; */
	
	/* get actual time & date
	newdt.time= Tgettime();
	newdt.date= Tgetdate();
	*/

	if(locate_a_file(file, file->env))
		/* can't find file */
		return 1;
	else
	{	/* touch it, touch me now : hard way
		h= Fopen(file->fname, FO_RW);
		Fdatime(&newdt, h, 1);
		Fclose(h);
		*/

		/* clever touch */
		set_time(&file->date, Tgettime()); /* get actual time */
		set_date(&file->date, Tgetdate()); /* get actual date */

		return 0;
	}
}


int toucher(File *file)
{
	int status= 0;
	File *f;
	
	if(!file) return 0;

	if(!file->cmd && !file->dependency)
		status|= touch_file(file);
	
	for(f= file->flist; f; f= f->next)
		status|= toucher(f);

	return status;	
}

int locator(File *file)
{
	int status= 1;
	File *f;
	
	if(!file) return 0;
	
	if(strcmp(file->fname, Make_Buildfilespec))
	{
		for(; file; file= file->next)
			for(f= file->flist; f; f= f->next)
				status&= locator(f);

		return status;
	}
	else
	{
		if(file->cmd)
		{
			/* efface l'objet */
			status= locate_a_file(file, file->env)==FATAL;
			status|= Fdelete(file->fname) <0;

			/* touch all source files, force la construction du projet */ 
			if(!status) 
				status= toucher(file);
		}
		else
			/* touch source file, force la construction du projet */ 
			touch_file(file);
			
		return 0;	
	}
}

int init_locator(File *prj)
{
	if(locator(prj))
		return FATAL;
	else
		return builder(prj);
}


void init_make(char *prj_fname)
{
	char *path;

	/* construit le chemin d'acces du repertoire courant */
	path= Path;
	*path++= Dgetdrv()+'A'; *path++= ':';
	Dgetpath(path, Dgetdrv() +1);
	path= &Path[strlen(Path)];
	*path++= '\\'; *path++= 0;
	
	/* tout en minuscules */
	strlwr(Path);
	strlwr(prj_fname);
	
	/* recupere la date complete du makefile */	
	if(Fsfirst(prj_fname, 0) <0)
		alert("fmake: can't open projectfile `%s`\n", prj_fname);

	set_time(&Make_Date, Dta.d_time);
	set_date(&Make_Date, Dta.d_date);

	printf("Makefile: %s\n", prj_fname);	
	printf("Current dir:\n%s\n\n", Path);
}

int do_it(char *prj_fname, int (*exec)(File *))
{
	int prj_file, status;
	File *project= NULL;

	/* ouvre le makefile */	
	init_make(prj_fname);
	prj_file= (int) Fopen(prj_fname, FO_READ);

	/* makefile parsing ... */
	init_parser(prj_file);
	parser(&project);
	
	Fclose(prj_file);	

	/* recupere l'environnement du tos */	
	get_tosenv();

	/* construit le projet */
	status= exec(project);

	/* libere les structures */
	free_file(project);

	/* cassos */	
	return status;
}

int make(char *prj_fname)
{
	return do_it(prj_fname, builder);
}

int make_clean(char *prj_fname)
{
	return do_it(prj_fname, cleaner);
}

int make_file(char *prj_fname)
{
	if(do_it(prj_fname, init_locator)==FATAL)
	{
		printf("fmake: can't find file `%s`\n", Make_Buildfilespec);
		return FATAL;
	}
	else
		return 0;
}

