/*
	fmake.h

	global defines
	
	T.AL 95	
*/

#if !defined(FMAKE_H)
#define FMAKE_H

#include <tos.h>

/* descriptin d'une variable */
typedef struct _var
{
	char *name;
	char *value;
	char *ext;
	
	struct _var *next;
} VAR;

/* description d'un argument de commande */
typedef struct _arg
{
	char *name;
	char *value;

	struct _arg *next;
} ARG;

/* description d'une commande */
typedef struct _cmd
{
	char *fname;
	
	char *inext;
	char *outext;
	ARG *argv;
	
	struct _cmd *next;
} CMD;

typedef union
{
	struct
	{
		unsigned h   : 5;	/* bit 15 */
		unsigned mn  : 6;
		unsigned sec : 5;	/* bit 0 */
	} tos_time;
	
	unsigned int int_time;
} TIME;

typedef union 
{
	struct
	{
		unsigned year  : 7;	/* bit 15 */
		unsigned month : 4;
		unsigned day   : 5;	/* bit 0 */
	} tos_date;
	
	unsigned int int_date;
} DATE;

typedef struct
{
	int year;
	int month;
	int day;

	int h;
	int mn;
	int sec;
} File_Date;

typedef union
{
	int date_array[6];
	DATE date; 
} DATE_ARRAY;

/* description d'un fichier */
typedef struct _file
{
	char *fname;
	File_Date date;
	int dependency;
	CMD *cmd;
	VAR *env;
	
	struct _file *flist;
	struct _file *next;
} File;


/* valeurs de retour */
#define FATAL 1
#define UPDATE -1
#define UP_TO_DATE 0

#define STOP_NEVER 0
#define STOP_FIRST 1

extern char Path[];
extern DTA Dta;
extern File_Date Make_Date;
extern VAR *tosenv;

extern int Make_Stop, Make_All;
extern char Make_Logpath[], Make_Logbasename[], Make_Logext[];
extern int Make_Logsinglefile;
extern char Make_Buildfilespec[];

/* functions prototypes */

/* parser */
int parse_a_cmd(File *file, char **token, char *outext, char *name);
int parse_a_var(File *file, char **token, char *name);
int parse_a_path(File *file, char **token, char *outext);
int parse_an_env(File *file, char **token);

int parse_a_fname(File **file, char **token);
int parse_a_acclist(File **file, char **token, CMD *cmd, VAR *env);
int parse_a_parlist(File **file, char **token, CMD *cmd, VAR *env);
int parse_a_file(File **file, char **token, CMD *cmd, VAR *env);
int parser(File **prj);


/* builder */
extern char spaces[];

VAR *find_var(char *name, char *ext, VAR *var);
int locate_a_file(File *file, VAR *env);

int build_project(File *prj);
int build_target(File *target, File *source);
int builder(File *prj);


/* punisher */
int exec_cmd(CMD *cmd, File *target, File *source);


/* main */
int make(char *prj_fname);
int make_clean(char *prj_fname);
int make_file(char *prj_fname);


#endif
