/*
	makefile parser
	
	T.AL 95
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined(FMAKE_H)
#include "fmake.h"
#endif
#if !defined(TOOLS_H)
#include "tools.h"
#endif

void check_cmd(CMD *cmd)
{
	int io_redir= 1, filespec;
	ARG *arg, *tmp;
	
	filespec= *Make_Logext;
				
	for(arg= cmd->argv; arg; arg= arg->next)
	{
		if(arg->name[0]=='>')
		{
			if(arg->name[1]==0)
			{
				io_redir= 0;
				tmp= arg->next;
				if(!arg->value && tmp && is_fname(tmp->name))
				{
					if(tmp->value)
					{
						free(arg->name);
						arg->name= strcatdup(">", tmp->name);
						arg->value= strdup(tmp->value);
					}
					else
						arg->value= strdup(tmp->name);
	
					/* detruit l'element de la chaine */
					tmp= tmp->next;
					free(arg->name);
					free(arg->value);
					arg->next= tmp;
				}
			}

			/* filespec patch */
			if(filespec)
			{
				free(arg->name);
				free(arg->value);
				if(Make_Logsinglefile)
				{
					arg->name= vstrcatdup(">%s%s%s", Make_Logpath, Make_Logbasename, Make_Logext);
					arg->value= NULL;
				}
				else
				{
					arg->name= strcatdup(">", Make_Logpath);
					arg->value= strcatdup(Make_Logbasename, Make_Logext);
				}
			}
		}
	}
	
	if(io_redir && filespec)
	{
		/* ajoute la redirection */
		arg= add_arg(cmd, new_arg());
		
		if(Make_Logsinglefile)
		{
			arg->name= vstrcatdup(">%s%s%s", Make_Logpath, Make_Logbasename, Make_Logext);
			arg->value= NULL;
		}
		else
		{
			arg->name= strcatdup(">", Make_Logpath);
			arg->value= strcatdup(Make_Logbasename, Make_Logext);
		}
	}
}

int parse_a_cmd(File *file, char **token, char *outext, char *name)
{
	char *tok;
	CMD *cmd= add_cmd(file, new_cmd());
	ARG *arg;

	/* lire ext */
	cmd->outext= outext;

	if(name) tok= name;
	else     tok= *token;
	
	/* verifier le type de tok : ext | fname */
	if(is_ext(tok))
	{
		cmd->inext= tok;
		/* lire la commande */
		cmd->fname= get_crtoken();
	}

	else if(is_fname(tok))
		cmd->fname= tok;

	else
		alert("fmake parser error 6: unexpected token %s\n", tok);
		
	if(name) tok= *token;
	else     tok= get_crtoken();
	
	while(tok && !is_cr(tok))
	{
		if(is_option(tok))
		{
			/* lire option */
			arg= add_arg(cmd, new_arg());
			
			arg->name= tok;
			if(tok= strchr(arg->name, '#'))
			{
				arg->value= strdup(tok);
				*tok= 0;
			}
			
			tok= get_crtoken();
		}
		else if(is_var(tok))
		{
			/* lire variable */
			arg= add_arg(cmd, new_arg());
			arg->name= tok;
			
			tok= get_crtoken();			
		}
		else
		{
			alert("fmake parser error 5: unexpected token %s\n", tok);
		}
	}

	check_cmd(cmd);

	*token= get_token();
	return tok!=NULL;
}

int parse_a_var(File *file, char **token, char *name)
{
	VAR *var= add_var(file, new_var());
	char *tok;
	char *chr;
	
	if(name) tok= name;
	else     tok= *token;
		
	/* lire var */
	var->name= tok;

	/* recherche le = */
	if(chr= strchr(tok, '=')) 
	{
		*chr= 0;	/*  la fin du nom de la variable */
		if(chr[1])
			tok= chr+1;
		else
			tok= get_token();
	}	
	else
	{
		if(name) tok= *token;
		else     tok= get_token();

		if(!is_a_equal(tok[0]))
			alert("fmake parser error 7: '=' expected\n");
				
		if(tok[1]) 
			strcpy(tok, tok+1);	/* au debut de la valeur de la variable */
		else
			tok= get_token();	/* non, le = est isole */
	}
	
	/* lire valeur */
	var->value= tok;
	
	*token= tok= get_token();
	return tok!=NULL;
}

int parse_a_path(File *file, char **token, char *outext)
{
	VAR *var= add_var(file, new_var());
	char *tok= *token;
	char *chr;

	/* lire var */
	var->ext= outext;
	var->name= tok;
		
	/* recherche le = */
	if(chr= strchr(tok, '=')) 
	{
		*chr= 0;	/*  la fin du nom de la variable */
		if(chr[1])
			tok= chr+1;
		else
			tok= get_token();
	}
	else
	{
		tok= get_token();
		if(!is_a_equal(tok[0]))
			alert("fmake parser error 7: '=' expected\n");
				
		if(tok[1]) 
			strcpy(tok, tok+1);	/* au debut de la valeur de la variable */
		else
			tok= get_token();	/* non, le = est isole */
	}
	
	/* lire valeur */
	var->value= tok;
	
	*token= tok= get_token();
	return tok!=NULL;
}

int parse_an_env(File *file, char **token)
{
	char *tok= *token;
	char *outext, *var;
	
	while(tok && !is_equal(tok))
	{
		if(is_ext(tok))	/* cmd  |  path */
		{
			/* lire outext */
			outext= tok;
				
			tok= get_token();
			if(is_ext(tok))	/* cmd */
				parse_a_cmd(file, &tok, outext, NULL);

			else if(is_var(tok))	/* path */
				parse_a_path(file, &tok, outext);
		}
		
		else if(is_var(tok))	/* variable | cmd */
		{
			var= tok;
			if(strchr(var, '='))
				parse_a_var(file, &tok, var);
			else
			{
				tok= get_token();
				
				if(is_a_equal(*tok))
					parse_a_var(file, &tok, var);
				else
					parse_a_cmd(file, &tok, NULL, var);
			}
		}
		
		else if(is_fname(tok))	/* project cmd */
			parse_a_cmd(file, &tok, NULL, NULL);
		
		else
			alert("fmake parser error 4: unexpected token %s\n", tok);
	}

	*token= tok;
	return tok!=NULL;
}


int parse_a_fname(File **file, char **token)
{
	char *tok= *token;

	*file= new_file();
	(*file)->fname= tok;
	
	*token= tok= get_token();
	return tok!=NULL;
}

int parse_a_acclist(File **file, char **token, CMD *cmd, VAR *env)
{
	File *f= NULL;
	char *tok= get_token();
	
	if(!*file)	*file= new_file();
	parse_an_env(*file, &tok);

	/* si l'environnement n'est pas defini -> heritage */
	if(!(*file)->env) (*file)->env= env;
	if(!(*file)->cmd) (*file)->cmd= cmd;

	tok= get_token();		/* = */

	while(tok && !is_clsacc(tok))
	{
		if(is_fname(tok))	/* project */
		{
			parse_a_file(&f, &tok, (*file)->cmd, (*file)->env);
			append_flist(file, f);
		}
		else
			alert("fmake parser error 3: unexpected token %s\n", tok);
	}
	
	*token= tok= get_token();
	return tok!=NULL;
}
		
int parse_a_parlist(File **file, char **token, CMD *cmd, VAR *env)
{
	File *f= NULL;
	char *tok= get_token();
	
	if(!*file)
	{
		*file= new_file();
		(*file)->env= env;
	}

	while(tok && !is_clspar(tok))
	{
		if(is_fname(tok))
		{
			parse_a_file(&f, &tok, cmd, env);
			append_flist(file, f);

			/* les dependances heritent de l'env du pere */
			for(; f; f= f->next)
				f->dependency= 1;
				
			/* f->env= env; */
		}
		else
		{
			alert("fmake parser error 2: unexpected token %s\n", tok);
		}
	}
	
	*token= tok= get_token();
	return tok!=NULL;
}

int parse_a_file(File **file, char **token, CMD *cmd, VAR *env)
{
	File *f= NULL;
	char *tok= *token;
	
	while(tok)
	{
		if(is_fname(tok))	/* project */
		{
			parse_a_fname(&f, &tok);
			add_file(file, f);
			f->env= env;
		}

		else if(is_opnacc(tok)) /* { */
		{
			parse_a_acclist(&f, &tok, cmd, env);
		}

		else if(is_opnpar(tok)) /* ( */
		{
			parse_a_parlist(&f, &tok, f && f->cmd ? f->cmd : cmd,
			                          f && f->env ? f->env : env);
		}

		else
			break;
	}

	*token= tok;
	return tok!=NULL;
}

int parser(File **file)
{
	int exit= 1;
	File *f= NULL;
	char *tok= get_token();

	while(exit)
	{
		exit= parse_a_file(&f, &tok, NULL, NULL);
		add_file(file, f);
	}
	
	return 0;
}
