/*
	tools.c
	
	some tool functions
	
	T.AL 95
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <tos.h>
#include <ext.h>

#if !defined(FMAKE_H)
#include "fmake.h"
#endif
#if !defined(TOOLS_H)
#include "tools.h"
#endif

int handle;
char buf[MAXBUF];
unsigned char c;
int (*is_type)(char);
int buflen;
char *ptr;
int inchar;


/* affiche un message et quiite */
void alert(char *msg, ...)
{
	va_list argptr;
	
	va_start(argptr, msg);
	vprintf(msg, argptr);
	
	puts("\nPress a key ...");
	
	getch();
	exit(1);
}


/* cree et vide une structure CMD */
CMD *new_cmd(void)
{
	CMD *new= (CMD *)malloc(sizeof(CMD));
	if(!new)
		alert("fmake: memory error in new_cmd\n");

	new->fname= NULL;
	new->argv= NULL;
	new->inext= NULL;
	new->outext= NULL;
	new->next= NULL;
	
	return new;
}

/* libere une liste de structure CMD */
int free_cmd(CMD *cmd)
{
	CMD *tmp;
	
	while(cmd)
	{
		tmp= cmd;
		cmd= cmd->next;
		
		free(tmp->fname);
		free_arg(tmp->argv);
		free(tmp->inext);
		free(tmp->outext);
		free(tmp);
	}
	
	return 0;
}

CMD *add_cmd(File *file, CMD *new)
{
	register CMD *tmp;
	
	for(tmp= file->cmd; tmp && tmp->next; tmp= tmp->next) ;

	if(tmp)
		tmp->next= new;
	else
		file->cmd= new;
	
	return new;
}


/* cree et vide une structure File */
File *new_file(void)
{
	File *new= (File *)malloc(sizeof(File));
	if(!new)
		alert("fmake: memory error in new_file\n");

	new->fname= NULL;
	new->date.sec= 0;
	new->date.mn= 0;
	new->date.h= 0;
	new->date.day= 0;
	new->date.month= 0;
	new->date.year= 0;
	new->dependency= 0;
	new->cmd= NULL;
	new->env= NULL;
	new->flist= NULL;
	new->next= NULL;

	return new;
}

/* libere une liste de structures File */
int free_file(File *file)
{
	File *tmp;
	
	while(file)
	{
		tmp= file;
		file= file->next;
		free(tmp->fname);
		free_var(tmp->env);
		free_cmd(tmp->cmd);
		free_file(tmp->flist);
		free(tmp);
	}
	
	return 0;
}

File *append_flist(File **file, File *new)
{
	register File *tmp;

	if(*file)
	{
		for(tmp= (*file)->flist; tmp && tmp->next; tmp= tmp->next) ;
		if(tmp)
			tmp->next= new;
		else
			(*file)->flist= new;
	}
	else
		*file= new;
	
	return new;
}

File *add_file(File **file, File *new)
{
	register File *tmp;
	
	if(*file)
	{
		for(tmp= *file; tmp->next; tmp= tmp->next) ;
		tmp->next= new;
	}
	else
		*file= new;
	
	return new;
}


ARG *new_arg(void)
{
	ARG *new= (ARG *)malloc(sizeof(ARG));
	if(!new)
		alert("fmake: memory error in new_arg\n");

	new->name= NULL;
	new->value= NULL;
	new->next= NULL;

	return new;
}

int free_arg(ARG *arg)
{
	ARG *tmp;
	
	while(arg)
	{
		tmp= arg->next;
		free(arg->name);
		free(arg->value);
		arg= tmp;
	}
	
	return 0;
}

ARG *add_an_arg(ARG **argv, ARG *new)
{
	register ARG *tmp;
	
	for(tmp= *argv; tmp && tmp->next; tmp= tmp->next) ;

	if(tmp)
		tmp->next= new;
	else
		*argv= new;

	return new;
}

ARG *add_arg(CMD *cmd, ARG *new)
{
	register ARG *tmp;
	
	for(tmp= cmd->argv; tmp && tmp->next; tmp= tmp->next) ;

	if(tmp)
		tmp->next= new;
	else
		cmd->argv= new;

	return new;
}

VAR *new_var(void)
{
	VAR *new= (VAR *)malloc(sizeof(VAR));
	if(!new)
		alert("fmake: memory error in new_var\n");

	new->name= NULL;
	new->value= NULL;
	new->ext= NULL;
	new->next= NULL;

	return new;
}

int free_var(VAR *var)
{
	VAR *tmp;
	
	while(var)
	{
		tmp= var->next;
		free(var->name);
		free(var->value);
		free(var->ext);
		var= tmp;
	}
	
	return 0;
}

VAR *add_var(File *file, VAR *new)
{
	register VAR *tmp;
	
	for(tmp= file->env; tmp && tmp->next; tmp= tmp->next) ;

	if(tmp)
		tmp->next= new;
	else
		file->env= new;

	return new;
}

VAR *add_a_var(VAR **env, VAR *new)
{
	register VAR *tmp;
	
	for(tmp= *env; tmp && tmp->next; tmp= tmp->next) ;

	if(tmp)
		tmp->next= new;
	else
		*env= new;

	return new;
}

VAR **add_an_env(VAR *old, VAR *new)
{
	register VAR *tmp;
	
	if(!old)	return NULL;
	
	for(tmp= old; tmp && tmp->next; tmp= tmp->next) ;

	if(tmp)
	{
		tmp->next= new;
		return &tmp->next;
	}

	return NULL;
}


int datecmp(File_Date *_a, File_Date *_b)
{
	int i;
	int *a, *b;

	a= (int *)_a;
	b= (int *)_b;
	for(i= 0; i<6 && *a==*b; a++, b++, i++) ;
	
	if(i==6)
		return 0;
	else if(*a>*b)
		return 1;
	else if(*a<*b)
		return -1;
	else
		alert("fmake: invalid date format in datecmp\n");		

	return 0;
}

File_Date *datecpy(File_Date *dst, File_Date *src)
{
	int i;
	int *a, *b;

	a= (int *)dst;
	b= (int *)src;
	for(i= 0; i<6 ; i++) *a++= *b++;
	
	return dst;	
}

File_Date *datemax(File_Date *_a, File_Date *_b)
{
	int i;
	int *a, *b;

	a= (int *)_a;
	b= (int *)_b;
	for(i= 0; i<6 && *a==*b; a++, b++, i++) ;
	
	if(i==6)
		return _a;
	else if(*a>*b)
		return _a;
	else if(*a<*b)
		return _b;
	else
		alert("fmake: invalid date format in datemax\n");		

	return 0;
}


void set_time(File_Date *time, unsigned int int_ftime)
{
	TIME ftime;
	ftime.int_time= int_ftime;
	
	time->sec= ftime.tos_time.sec<<1;
	time->mn=  ftime.tos_time.mn;
	time->h=   ftime.tos_time.h;
}

void set_date(File_Date *date, unsigned int int_fdate)
{
	DATE fdate;
	fdate.int_date= int_fdate;
	
	date->day=   fdate.tos_date.day;
	date->month= fdate.tos_date.month;
	date->year=  fdate.tos_date.year + 1980;
}

void set_file_dtm(File *file)
{
	DTA *dta;

	if(!file->date.year)
	{
		Fsfirst(file->fname, 0);	
		dta= Fgetdta();
		set_time(&file->date, dta->d_time);
		set_date(&file->date, dta->d_date);
	}
}

int init_parser(int prj_file)
{	
	handle= prj_file;
	ptr= NULL;
	inchar= -1;
	buflen= 0;
	c= read_a_char();

	return 0;
}

int read_a_char(void)
{
	if(inchar==buflen-1)
	{
		buflen= (int) Fread(handle, MAXBUF, buf);
		ptr= buf;
		inchar= 0;
		if(buflen == 0) return 0;
	}
	else
	{
		inchar++;
		ptr++;
	}
		
	return *ptr;
}

char *get_token(void)
{
	char *tok;
	char token[MAXBUF];

	if(!c) return NULL;
	while(c && is_a_crspace(c)) c= read_a_char();
	
	tok= token;
	is_type= find_type(c);
	if(is_type == is_a_separator)
	{
		*tok++= c;
		c= read_a_char();
	}
	else if(is_type == is_a_str)
	{
		c= read_a_char(); /* saute le premier \" */
		while(c && !is_a_str(c))
		{
			*tok++= c;
			c= read_a_char();
		}
		c= read_a_char(); /* saute le dernier \" */
	}
	else
	{
		while(c && is_type(c))
		{
			*tok++= c;
			c= read_a_char();
		}
	}
	*tok= 0;
	
	while(c && is_a_crspace(c)) c= read_a_char();
	
	return strdup(token);
}

char *get_crtoken(void)
{
	char *tok;
	char token[MAXBUF];

	if(!c) return NULL;

	while(c && is_a_space(c)) c= read_a_char();
	
	tok= token;
	is_type= find_crtype(c);
	if(is_type == is_a_separator)
	{
		*tok++= c;
		c= read_a_char();
	}
	else if(is_type == is_a_str)
	{
		c= read_a_char(); /* saute le premier \" */
		while(c && !is_a_str(c))
		{
			*tok++= c;
			c= read_a_char();
		}
		c= read_a_char(); /* saute le dernier \" */
	}
	else
	{
		while(c && is_type(c))
		{
			*tok++= c;
			c= read_a_char();
		}
	}
	*tok= 0;
	
	while(c && is_a_space(c)) c= read_a_char();
	
	return strdup(token);
}


int is_cr(char *tok)
{
	return is_a_cr(*tok);
}

int is_fname(char *tok)
{
	char *tmp;
	
	if(!tok) return 0;
	
	if(*tok=='\.' || !*tok) return 0;
	for(tmp= tok; *tmp && is_a_alpha(*tmp); tmp++) ;

	return !*tmp;
}

int is_ext(char *tok)
{
	int len;

	if(!tok) return 0;
	
	len= (int) strlen(tok);
	if(len>4) return 0;
	if(*tok!='\.') return 0;
	
	return 1;
}

int is_option(char *tok)
{
	char *tmp;
	
	if(!tok || *tok=='#') return 0;
		
	for(tmp= tok; 
	    *tmp && is_a_option(*tmp++);
	   ) ;

	return !*tmp;
}

int is_var(char *tok)
{
	if(!tok) return 0;
	
	return *tok=='\#';
}

int is_equal(char *tok)
{
	if(!tok) return 0;

	return *tok=='=' && tok[1]==0;
}
int is_opnacc(char *tok)
{
	if(!tok) return 0;

	return *tok=='{';
}

int is_clsacc(char *tok)
{
	if(!tok) return 0;

	return *tok=='}';
}

int is_opnpar(char *tok)
{
	if(!tok) return 0;

	return *tok=='(';
}

int is_clspar(char *tok)
{
	if(!tok) return 0;

	return *tok==')';
}

int (*find_type(char c))(char)
{
	int (*type)(char);

	if(is_a_crspace(c))
		type= is_a_crspace;

	else if(is_a_str(c))
		type= is_a_str;

	else if(is_a_alpha(c))
		type= is_a_alpha;

	else if(is_a_separator(c))
		type= is_a_separator;

	else
		alert("fmake parser error: character %c (%d) not recognised\n", c, c);

	return type;
}

int (*find_crtype(char c))(char)
{
	int (*type)(char);

	if(is_a_space(c))
		type= is_a_space;

	else if(is_a_str(c))
		type= is_a_str;
		
	else if(is_a_alpha(c))
		type= is_a_alpha;

	else if(is_a_cr(c))
		type= is_a_cr;

	else if(is_a_separator(c))
		type= is_a_separator;

	else
		alert("parser error: character %c (%d) not recognised\n", c, c);
	
	return type;
}


int is_a_char(char c)
{
	return c!=0; /* of course */
}

int is_a_space(char c)
{
	return c==' ' 
	    || c==9;
}

int is_a_crspace(char c)
{
	return c==' ' 
	    || c=='\n'
	    || c=='\r'
	    || c==9;
}

int is_a_str(char c)
{
	return c=='\"';
}

int is_a_cr(char c)
{
	return c=='\n' 
	    || c=='\r';
}

int is_a_equal(char c)
{
	return c=='=';
}

int is_a_alpha(char c)
{
	return c>='A' && c<='z' 
	    || c>='0' && c<='9' 
	    || c=='_' 
	    || c=='-' 
	    || c=='+'
	    || c=='/'
	    || c=='.' 
	    || c==':' 
	    || c=='\\'
	    || c=='#'
	    || c=='='
	    || c=='>';
}

int is_a_option(char c)
{
	return c>='A' && c<='z' 
	    || c>='0' && c<='9' 
	    || c=='_' 
	    || c=='-' 
	    || c=='+'
	    || c=='/'
	    || c=='.' 
	    || c==':' 
	    || c=='\\'
	    || c=='#'
		|| c=='='
	    || c=='>';
}


int is_a_separator(char c)
{
	return c=='(' 
	    || c=='{' 
	    || c=='}' 
	    || c==')' 
	    || c=='\,' 
	    || c=='=';
}


void get_tosenv(void)
{
	char *tmp;
	char *env= _BasPag->p_env;
	VAR *var;
	
	while(*env)
	{
		var= add_a_var(&tosenv, new_var());

		/* ajoute un # au debut du nom de la varible */
		var->name= malloc(strlen(env) +2);
		var->name[0]='#';
		strcpy(&var->name[1], env);
		
		/* recupere la valeur */
		tmp= strchr(var->name, '=');
		if(tmp)	*tmp++= 0;
		var->value= tmp;

		/* suivante */
		env= env + strlen(env) +1;
	} 
}


char *basename(char *path)
{
	char *tmp= strrchr(path, '\\');
	
	if(tmp) return tmp +1;
	else    return path;
}

char *getext(char *fname)
{
	return strchr(basename(fname), '.');
}

char *set_ext(char *ext, char *fname)
{
	char *tmp= strchr(fname, '.');
	
	if(!tmp)
		strcat(fname, ext);
	else
		strcpy(tmp, ext);
		
	return fname;
}

void splitfilename(char *fname, char **path, char **bname, char **ext)
{
	char *str, *base;
	
	base= basename(fname);
	*bname= strdup(base);
	str= strchr(*bname, '.');
	if(!str) str= "";
	*ext= strdup(str);
	*str= 0;
	*path= strndup(fname, base - fname);
}

char *filename(char *path, char *bname, char *ext)
{
	char *dup= malloc(strlen(path) + strlen(bname) + strlen(ext));
	
	if(!dup)
		alert("fmake: memory error in create filename\n");
	
	else
	{
		strcpy(dup, path);
		strcat(dup, bname);
		strcat(dup, ext);
	}
	return dup;
}

char *strvar(char *str)
{
	return strchr(str, '#');
}

char *strcatdup(char *str1, char *str2)
{
	char *dup= malloc(strlen(str1) + strlen(str2));
	
	if(!dup)
		alert("fmake: memory error in strcatdup\n");
	
	else
	{
		strcpy(dup, str1);
		strcat(dup, str2);
	}
	return dup;
}

char *vstrcatdup(char *format, ...)
{
	long len;
	char *str, *dup, *dst, *aux;
	va_list a_str, a_base;
	
	va_start(a_str, format);
	a_base= a_str;
	
	len= 0;
	str= format;
	while(*str)
	{
		if(*str!='%')
			len++;

		else
		{
			len+= strlen( (char *) va_arg(a_str, char *) );
			if(*(++str)!='s')
				alert("fmake: vstrcatdup format error\n");
		}
		str++;
	}
	
	dup= malloc(len +1L);
	if(!dup) alert("fmake: memory error in vstrcatdup\n");
		
	a_str= a_base;
	str= format;
	dst= dup;
	while(*str)
	{
		if(*str=='%')
		{
			aux= va_arg(a_str, char *);
			while(*aux) *dst++= *aux++;
			
			if(*(++str)=='s')
				str++;
			else
				alert("fmake: vstrcatdup format error\n");
		}
		else
			*dst++= *str++;
	}
	va_end(a_str);
	*dst= 0;
	
	return dup;
}

char *strndup(char *str, int n)
{
	char *dup= malloc((long)n +1);
	
	if(!dup)
		alert("fmake: memory error in strndup\n");
	
	else
	{
		strncpy(dup, str, n);
		dup[n]= 0;
	}
	return dup;
}

